//
// THIS FILE IS COPYRIGHT 1994, 1995 BY TELENOR R&D, NORWAY
//
//                        ALL RIGHTS RESERVED.
//
// For rights granted to the public see the COPYRIGHT file
//
//
// Disk.h	Handles all files on disk. Writes out words to index
//		files, merges index files and searches index files.
//
// Author:  Ken Ronny Schouten
//
// Changes: January 1995 Brd Hfjeld
//				Added some new functions during restructuring
//

#ifndef _DISK_H
#define _DISK_H

#include <stream.h>
#include <fstream.h>
#include <streambuf.h>
#include <stdio.h>
#include "Entry.h"

class Index;

// The _mode values
#define DREAD  1
#define DWRITE 2

// _error values
#define DOK   	0		// No error
#define DMODE 	1
#define DOPEN 	2		// Error opening file
#define DEOF  	3		// End of file
#define DBIG  	4		// An entry was too big to fit into index file
#define DBAD  	5		// Bad index, header inconsistent
#define DIO   	6		// IO error
#define DOVERFLOW 7		// Internal table overflow
#define DNODIX	8		// Can't search for '*', no date index
#define DNOTHAPPEN 9		// Internal error, this could not happen

// The index file block size
#define BLOCK_SIZE  8192   	// 8K of info stored in one block

struct search_params {
    long after;			// Documents dated after this date
    long before;		// Documents dated before this date	
    long max_internal;		// Max size of internal url lists
    long max_output;		// Max number of urls to output
};

class Disk {
public:
    friend class Dictionary;

    // Constructor and Destructor
    Disk(unsigned char *s, int mode);
    ~Disk();

    // Logical entry access methods
    Entry* nextEntry(int initialize);
                                // Read next entry from disk
    int nextFrequency(unsigned char* & word, int initialize);
                                // Read next frequency entry from disk

    // Writing indexes on disk
    int MakeIndexOnDisk(Index& index);

    // Find (truncated) words in index
    URLPTRSplaySet *Search(unsigned char *word, struct search_params *sp);

    // Merge two indexes into one
    int Merge(Disk& d1, Disk& d2);

    // Information about index
    int blocks();               // Number of blocks in index
    int error();                // Last error
    unsigned char *Name();	// Index name

    // Index file access methods, updates _buffer, _ind and _last
    int GetIndexBlock(long blockno);
                                // Get a given block from index file
    int GetNextIndexBlock();	// Read next block from index file
    int GetNextIndexEntry(unsigned char **word, unsigned long *offset);
                                // Read next index entry, load new block if req
    int SetNextIndexEntry(unsigned char *word, unsigned long offset);
                                // Append an index entry to index file
    int FlushBuffer();		// Flush last block

    // Pointer file access
    int GetUrls(URLPTRSplaySet *urllist, long offset, 
		struct search_params *sp);
                                // Load Url list conditionally
    int GetNumberOfPointers(unsigned long offset);
                                // Get number of entries and set file pointer
    // URL file access
    int GetURLRecord(unsigned offset, long *date, unsigned char *urlname,
		      unsigned char *title);
    int GetNextURLRecord(long *date, unsigned char *urlname,
		      unsigned char *title);

    // Date index file access
    int GetDixEntry(unsigned long dix_offset, long *date, 
		    unsigned long *url_offset);
    int GetNextDixEntry(long *date, unsigned long *url_offset);
    int SetNextDixEntry(long date, unsigned long url_offset);


private:
    // The index header
    struct {
	long   blocks;
	long   bsize;
	int    letter[256];
    } _header;

    // File pointers
    FILE *_fp_ind;
    FILE *_fp_ptr;
    FILE *_fp_url;
    FILE *_fp_dix;

    // Disk status
    int  _mode;
    int  _error;
    unsigned char *_indexname;

    // The index buffer
    unsigned char  *_buffer;   	// Block buffer
    int   _ind;		     	// Current index in buffer
    int   _block;		// Current block number in buffer
    int   _last;		// Last word in buffer

    // File pointer into pointer file
    unsigned long pointer_fp;

    int CopyPointers(int num1, FILE *file1, int num2, FILE *file2, 
		     unsigned long offset2);
                                // This beast concatenates two pointer lists
    // Storing entries while building index
    int StoreEntry(Entry& e);   // Store entry on disk

    // Store URL list to file, will also calculate offsets
    int StoreUrl(URLPTRSplaySet u);

    // Read/write header
    int StoreHeader();
    int ReadHeader();

    // Test for presence of word in a block
    int WordInBlock(unsigned char *word, int truncated); 

    // Search routines
    URLPTRSplaySet *SearchForWord(unsigned char *word, 
                                  struct search_params *sp);
    URLPTRSplaySet *SearchForDate(struct search_params *sp);

    // Build the date index
    int MakeDateIndex(Index &index);
};

inline int Disk::blocks()
{
    return _header.blocks; 
}

inline int Disk::error()
{
    return _error;
}

inline unsigned char *Disk::Name()
{
    return _indexname;
}

#endif /* _DISK_H */

